#!/usr/bin/env python
# project: ulws
# description: Ultra Light Web Server
# file: ulws/__init__.py
# file_version: 1.4
# author: DANA <dkovacs@deasys.eu>
# license: GPL 3.0
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

__version__ = "1.4"
#!/usr/bin/env python2.7

import sys, os
import types
import posixpath
import datetime, time
import traceback
import weakref
#import json
import cgi
import mimetypes
import urllib

from SocketServer import ThreadingMixIn
from BaseHTTPServer import BaseHTTPRequestHandler, HTTPServer
from urlparse import urlparse
from threading import local, Event
from datetime import datetime, timedelta

from sutils import *


# ---------------------------------------------------
# __all__
# ---------------------------------------------------

__all__ = qlist()


# ---------------------------------------------------
# ConfigurationError
# ---------------------------------------------------

@__all__.register
class ConfigurationError(Exception): pass
 

# ---------------------------------------------------
# guess_type()
# ---------------------------------------------------

def guess_type(path):
    base, ext = posixpath.splitext(path)
    if ext in _extensions_map:
        return _extensions_map[ext]
    ext = ext.lower()
    if ext in _extensions_map:
        return _extensions_map[ext]
    else:
        return _extensions_map['']

if not mimetypes.inited:
    mimetypes.init() # try to read system mime.types
_extensions_map = mimetypes.types_map.copy()
_extensions_map.update({
    '': 'text/plain', # Default
    '.py': 'text/plain',
    '.c': 'text/plain',
    '.h': 'text/plain',
    '.md': 'text/plain',
    '.template': 'text/plain',
})


# ---------------------------------------------------
# FieldStorage
# ---------------------------------------------------    

class FieldStorage(cgi.FieldStorage):
    
    def __init__(self, *args, **kwds):
        try:
            cgi.FieldStorage.__init__(self, *args, **kwds)
        except ValueError, ex:
            if str(ex) == 'Maximum content length exceeded':
                raise HttpErrorResponse(status=413)
            else:
                raise ex

    def read_lines_to_eof(self):
        """Internal: read lines until EOF."""
        while 1:
            line = self.fp.readline(1<<16)
            if not line:
                self.done = -1
                break
            self.__write(line)
    
    def read_lines_to_outerboundary(self):
        """Internal: read lines until outerboundary."""
        next = "--" + self.outerboundary
        last = next + "--"
        delim = ""
        last_line_lfend = True
        while 1:
            line = self.fp.readline(1<<16)
            if not line:
                self.done = -1
                break
            if line[:2] == "--" and last_line_lfend:
                strippedline = line.strip()
                if strippedline == next:
                    break
                if strippedline == last:
                    self.done = 1
                    break
            odelim = delim
            if line[-2:] == "\r\n":
                delim = "\r\n"
                line = line[:-2]
                last_line_lfend = True
            elif line[-1] == "\n":
                delim = "\n"
                line = line[:-1]
                last_line_lfend = True
            else:
                delim = ""
                last_line_lfend = False
            self.__write(odelim + line)
    
    def skip_lines(self):
        """Internal: skip lines until outer boundary if defined."""
        if not self.outerboundary or self.done:
            return
        next = "--" + self.outerboundary
        last = next + "--"
        last_line_lfend = True
        while 1:
            line = self.fp.readline(1<<16)
            if not line:
                self.done = -1
                break
            if line[:2] == "--" and last_line_lfend:
                strippedline = line.strip()
                if strippedline == next:
                    break
                if strippedline == last:
                    self.done = 1
                    break
            if line.endswith('\n'):
                last_line_lfend = True
            else:
                last_line_lfend = False

    
    def items(self):
        items = {}
        for key in self.keys():
            value_list = self[key]
            if isinstance(value_list, list):
                items[key] = []
                for item in value_list:
                    if item.filename is not None:
                        value = item
                    else:
                        value = item.value
                    items[key].append(value)
            else:
                if value_list.filename is not None:
                    value = value_list 
                else:
                    value = value_list.value
                items[key] = value
        return items  


# ---------------------------------------------------
# MultipartWrapper
# ---------------------------------------------------    

class MultipartWrapper(object):
    """Wraps a file-like object, returning '' when Content-Length is reached.
    
    The cgi module's logic for reading multipart MIME messages doesn't
    allow the parts to know when the Content-Length for the entire message
    has been reached, and doesn't allow for multipart-MIME messages that
    omit the trailing CRLF (Flash 8's FileReference.upload(url), for example,
    does this). The read_lines_to_outerboundary function gets stuck in a loop
    until the socket times out.
    
    This rfile wrapper simply monitors the incoming stream. When a read is
    attempted past the Content-Length, it returns an empty string rather
    than timing out (of course, if the last read *overlaps* the C-L, you'll
    get the last bit of data up to C-L, and then the next read will return
    an empty string).
    """
    
    def __init__(self, rfile, clen):
        self.rfile = rfile
        self.maxlen = rfile.maxlen
        self.clen = clen
        self.bytes_read = 0
    
    def read(self, size = None):
        #print "MultipartWrapper.read", size, self.clen, self.bytes_read
        if self.clen:
            # Return '' if we've read all the data.
            if self.bytes_read >= self.clen:
                return ''
            
            # Reduce 'size' if it's over our limit.
            new_bytes_read = self.bytes_read + size
            if new_bytes_read > self.clen:
                size = self.clen - self.bytes_read
        
        data = self.rfile.read(size)
        self.bytes_read += len(data)
        return data
    
    def readline(self, size = None):
        #print "MultipartWrapper.readline", size, self.clen, self.bytes_read
        if size is not None:
            if self.clen:
                # Return '' if we've read all the data.
                if self.bytes_read >= self.clen:
                    return ''
                
                # Reduce 'size' if it's over our limit.
                new_bytes_read = self.bytes_read + size
                if new_bytes_read > self.clen:
                    size = self.clen - self.bytes_read
            
            data = self.rfile.readline(size)
            self.bytes_read += len(data)
            return data
        
        # User didn't specify a size ...
        # We read the line in chunks to make sure it's not a 100MB line !
        res = []
        size = 256
        while True:
            if self.clen:
                # Return if we've read all the data.
                if self.bytes_read >= self.clen:
                    return ''.join(res)
                
                # Reduce 'size' if it's over our limit.
                new_bytes_read = self.bytes_read + size
                if new_bytes_read > self.clen:
                    size = self.clen - self.bytes_read
            
            data = self.rfile.readline(size)
            self.bytes_read += len(data)
            res.append(data)
            # See http://www.cherrypy.org/ticket/421
            if len(data) < size or data[-1:] == "\n":
                return ''.join(res)
    
    def readlines(self, sizehint = 0):
        # Shamelessly stolen from StringIO
        total = 0
        lines = []
        line = self.readline()
        while line:
            lines.append(line)
            total += len(line)
            if 0 < sizehint <= total:
                break
            line = self.readline()
        return lines
    
    def close(self):
        self.rfile.close()
    
    def __iter__(self):
        return self.rfile
    
    def next(self):
        #print "MultipartWrapper.next", self.clen, self.bytes_read
        if self.clen:
            # Return '' if we've read all the data.
            if self.bytes_read >= self.clen:
                return ''
        
        data = self.rfile.next()
        self.bytes_read += len(data)
        return data


# ---------------------------------------------------
# MaxSizeExceeded
# ---------------------------------------------------

class MaxSizeExceeded(Exception): pass

# ---------------------------------------------------
# SizeCheckWrapper
# ---------------------------------------------------

class SizeCheckWrapper(object):
    """Wraps a file-like object, raising MaxSizeExceeded if too large."""
    
    def __init__(self, rfile, maxlen):
        self.rfile = rfile
        self.maxlen = maxlen
        self.bytes_read = 0
    
    def _check_length(self):
        if self.maxlen and self.bytes_read > self.maxlen:
            raise MaxSizeExceeded()
    
    def read(self, size=None):
        data = self.rfile.read(size)
        self.bytes_read += len(data)
        self._check_length()
        return data
    
    def readline(self, size=None):
        if size is not None:
            data = self.rfile.readline(size)
            self.bytes_read += len(data)
            self._check_length()
            return data
        
        # User didn't specify a size ...
        # We read the line in chunks to make sure it's not a 100MB line !
        res = []
        while True:
            data = self.rfile.readline(256)
            self.bytes_read += len(data)
            self._check_length()
            res.append(data)
            # See http://www.cherrypy.org/ticket/421
            if len(data) < 256 or data[-1:] == "\n":
                return ''.join(res)
    
    def readlines(self, sizehint=0):
        # Shamelessly stolen from StringIO
        total = 0
        lines = []
        line = self.readline()
        while line:
            lines.append(line)
            total += len(line)
            if 0 < sizehint <= total:
                break
            line = self.readline()
        return lines
    
    def close(self):
        self.rfile.close()
    
    def __iter__(self):
        return self
    
    def next(self):
        data = self.rfile.next()
        self.bytes_read += len(data)
        self._check_length()
        return data


# ---------------------------------------------------
# HttpBaseResponse
# ---------------------------------------------------

@__all__.register
class HttpBaseResponse(object):

    code = 200
    footer = '<br><hr><small>Ultra Ligth Web Server {}</small>'.format(__version__)

    def __init__(self, response, content_type = None, content_length = None, code = None, headers = None, **quick_headers ):
        self.code = code or self.code
        self.headers = headers or qdict()
        for name in quick_headers:
            self.headers[name.replace('_','-')] = quick_headers[name]
        self.response = response
        if content_length:
            self.headers['Content-Length'] = content_length
        if response and not self.headers.get('Content-Length', None) and not isinstance(self.response, types.GeneratorType):
            self.content_length = len(self.response)
        if content_type:
            self.headers['Content-Type'] = content_type

    def __iter__(self):
        yield self.response


# ---------------------------------------------------
# HttpResponse
# ---------------------------------------------------

@__all__.register
class HttpResponse(HttpBaseResponse):

    default_content_type = 'text/html'

    def __init__( self, *args, **kwargs ):
        kwargs.setdefault('Content_Type', self.default_content_type)
        super(HttpResponse,self).__init__(*args, **kwargs)


# ---------------------------------------------------
# HttpGeneratorResponse
# ---------------------------------------------------

@__all__.register
class HttpGeneratorResponse(object):

    def __iter__(self):
        return self.response

# ---------------------------------------------------
# HttpFileResponse
# ---------------------------------------------------

@__all__.register
class HttpFileResponse(HttpBaseResponse):

    def __init__( self, path, filename = None ):
        super(HttpFileResponse,self).__init__(None, Content_Lenght = os.path.getsize(path) )
        self.path = path
        if filename:
            self.headers["Content-Disposition"] = u'inline; filename="{}";'.format( filename ).encode( 'utf-8' )

    def __iter__(self):
        buffer_size = 4096
        f = open( self.path, 'rb' )
        data = f.read(buffer_size)
        while data:
            yield data
            data = f.read(buffer_size)


# ---------------------------------------------------
# HttpRedirectResponse
# ---------------------------------------------------

@__all__.register
class HttpRedirectResponse( HttpBaseResponse ):

    code = 302
    page_template = "<html><head><title>Moved</title></head><body><h1>Moved</h1><p>This page has moved to <a href=\"{url}\">{url}</a>.</p></body></html>"

    def __init__( self, url, message = None, **options ):
        super(HttpRedirectResponse,self).__init__( message or self.page_template.format( url = url ), **options )
        self.headers['Location'] = url


# ---------------------------------------------------
# HttpErrorResponse
# ---------------------------------------------------

@__all__.register
class HttpErrorResponse(HttpBaseResponse, Exception):

    page_template = "<html><head><title>{code} {name}</title></head><body><h1>{name}</h1>{message}{traceback}{footer}</body></html>"
    traceback_template = "<hr><h2>Traceback</h2><pre>{traceback}</pre>"

    code = 500
    name = "Internal Server Error"
    message = None
    message_template = "An error has occured while handling request to asf URL: {url}<br><b><pre>{message}</pre></b>"

    def __init__(self, message = None, code = None, name = None, traceback = None, url = None, **options):
        code = code or self.code
        name = name or self.name
        message_template = options.pop('message_template', self.message_template) or '{}'
        message = message_template.format(code = code, name = name, message = (message or self.message), url = url or '' )
        options.setdefault('content_type', 'text/html')
        if traceback:
            traceback = self.traceback_template.format(traceback = traceback)
        else:
            traceback = ''
        message = self.page_template.format(code = code, name = name, message = message, traceback = traceback, footer = self.footer )
        super(HttpErrorResponse,self).__init__(message, **options)


# ---------------------------------------------------
# HttpBadRequestResponse
# ---------------------------------------------------

@__all__.register
class HttpBadRequestResponse(HttpErrorResponse):

    code = 400
    name = "Bad Request"
    message = None
    message_template = "An error occured while processing the request: {message}"


# ---------------------------------------------------
# HttpNotFoundResponse
# ---------------------------------------------------

@__all__.register
class HttpNotFoundResponse(HttpBadRequestResponse):

    code = 404
    name = "Not Found"
    message = None
    message_template = "The requested URL {message} was not found on this server."


# ---------------------------------------------------
# HttpMethodNotAllowedResponse
# ---------------------------------------------------

@__all__.register
class HttpMethodNotAllowedResponse(HttpBadRequestResponse):

    code = 405
    name = "Method Not Allowed"
    message = None
    message_template = "Method not allowed: {message}"


# ---------------------------------------------------
# HttpTimeOutResponse
# ---------------------------------------------------

@__all__.register
class HttpTimeOutResponse(HttpBadRequestResponse):

    code = 420  # ;-)
    name = 'Timeout'
    message = None
    message_template = "Request timed out: {message}"


# ---------------------------------------------------
# HttpRequestDispatcher
# ---------------------------------------------------

@logged
class HttpRequestDispatcher(BaseHTTPRequestHandler, object):

    def parse_body(self):
        #self.params = {}
        self.request_content_length = 0
        self.rfile = SizeCheckWrapper(self.rfile, 0)
        try:
            self.request_content_length = int(self.headers.get( "Content-Length", 0 ))
        except ValueError:
            pass
        #if not self.request_content_length:
        #    return
        self.rfile.maxlen = self.request_content_length
        self.request_content_type = self.headers.get( 'Content-Type', '' )
        #if not ( self.request_content_type.startswith( 'multipart/' ) or self.request_content_type.startswith( 'application/x-www-form' ) ):
        #    return
        self.environ = qdict(os.environ)
        self.environ.REQUEST_METHOD = self.command
        self.environ.HEADERS = self.headers
        self.environ.CONTENT_TYPE = self.request_content_type
        self.environ.CONTENT_LENGTH = self.request_content_length
        self.environ.QUERY_STRING = self.query_string
        self.environ.PATH = self.path
        self.rfile = MultipartWrapper( self.rfile, self.request_content_length )
        try:
            self.params = FieldStorage( fp = self.rfile, environ = self.environ, keep_blank_values = True ).items()
        except Exception, e:
            if e.__class__.__name__ == 'MaxSizeExceeded':
                raise self.send_error( 413 )
            raise

    def do_GET(self, **options ):
        self.dispatch_request()        

    def do_POST(self, **options):
        self.dispatch_request()

    def do_PUT(self, **options):
        self.dispatch_request()

    def do_DELETE(self, **options):
        self.dispatch_request()

    def do_HEAD(self, **options ):
        self.dispatch_request()

    def dispatch_request(self):
        scheme, location, self.path, self.params, self.query_string, frag = urlparse(self.path)
        self.parse_body()
        try:
            response = self.server.dispatch( self )
            if not isinstance( response, HttpBaseResponse ):
                response = HttpResponse( response )
            if isinstance(response, HttpErrorResponse):
                raise response
            if not response.headers.get('Content-Type', None):
                response.headers['Content-Type'] = guess_type(self.path)
        except HttpErrorResponse as exc:
            if not ( (exc.code >= 300) and (exc.code < 500) ):
                traceback.print_exc()
            response = exc
        except Exception as exc:
            traceback.print_exc()
            response = HttpErrorResponse( "{}: {}".format(exc.__class__.__name__, str(exc)), traceback = traceback.format_exc(), url = self.path )
        self.handle_response(response)

    def handle_response(self, response):
        try:
            self.send_response(response.code)
            for key, value in response.headers.iteritems():
                self.send_header( key, value )
            self.send_header('Connection', 'close')
            self.end_headers()
            if self.command == 'HEAD' or response.code < 200 or response.code in (204, 304):
                return
            for data in response:
                self.wfile.write(data)
        except Exception as exc:
            print( "Failed to transmit response!" )
            traceback.print_exc()

    def log_error(self, msg, *args):
        self.__logger.error("[{}] {}".format(self.client_address[0], msg % args ))

    def log_message(self, msg, *args):
        self.__logger.info("[{}] {}".format(self.client_address[0], msg % args ))


# ---------------------------------------------------
# Dispatcher
# ---------------------------------------------------

class Dispatcher(dict):

    _allow_quick_traverse = False

    def parse( self, path, *args, **kwargs ):
        return path

    def dispatch( self, path, *args, **kwargs ):
        node = self
        path = self.parse( path, *args, **kwargs )
        kwargs.setdefault('full_path', path)
        path_len = len(path)
        path_pos = 0
        while path_pos < path_len:
            lookfor = path[path_pos]
            if not lookfor in node:
                break
            node = self[lookfor]
            if not isinstance(node,Dispatcher):
                break
            if not node._allow_quick_traverse:
                return node.dispatch( path[path_pos+1:], *args, **kwargs )
            path_pos += 1
        return node( path[path_pos:], *args, **kwargs )

    def __call__( self, path, *args, **kwargs ):
        raise NotImplemented()


# ---------------------------------------------------
# expose
# ---------------------------------------------------

@__all__.register
def exposed( alias = None ):
    def _exposed(obj):
        obj.exposed = True
        if alias:
            obj.alias = alias
        return obj
    return _exposed


# ---------------------------------------------------
# Interface
# ---------------------------------------------------

@__all__.register
class Interface(Dispatcher):

    _allow_quick_traverse = False

    def parse( self, path, *args, **kwargs ):
        if isinstance(path, list):
            return path
        return path.split( '/' )[1:]

    def __call__( self, path, *args, **kwargs ):
        #print "interface.call", self, path, args, kwargs
        method = getattr(self, path[0], None )
        #print "interface.call", 'method 1: ', method
        if method and getattr(method, 'exposed', None):
            path = path[1:]
        else:
            method = self.default
        #print "interface.call", "method 2: ", method
        if len(args) == 0:
            args = path
        return method( *args, **kwargs )

    @exposed()
    def default( self, *path, **kwargs ):
        raise HttpNotFoundResponse( '/' + '/'.join( kwargs['full_path'][:-1] ) )


# ---------------------------------------------------
# MultiThreadedHTTPServer
# ---------------------------------------------------

class MultiThreadedHTTPServer(ThreadingMixIn, HTTPServer, object): pass


# ---------------------------------------------------
# Server
# ---------------------------------------------------

@__all__.register
@logged
class Server(MultiThreadedHTTPServer):

    def __init__(self, root_interface, server_address, request_handler_class = HttpRequestDispatcher, bind_and_activate=True):
        super(MultiThreadedHTTPServer,self).__init__(server_address, request_handler_class, bind_and_activate)
        self.root_interface = root_interface

    def dispatch( self, request ):
        #print "Dispatching request: {} {} Params: {}".format( request.command, request.path, request.params )
        path = urllib.unquote(request.path.rstrip('/')+'/')
        path = path + request.command.lower()
        kwargs = request.params
        kwargs['full_url'] = request.path
        return self.root_interface.dispatch( path, **kwargs )

    def serve_forever(self):
        self.address = self.socket.getsockname()
        self.__logger.info("Serving HTTP on {} port {}".format( *self.address ))
        self.__logger.info("Press Ctrl+C to exit.")
        try:
            super(Server,self).serve_forever()
        except KeyboardInterrupt:
            print "KeyboardInterrupt received, bye-bye"




# ---------------------------------------------------
# StaticFileInterface
# ---------------------------------------------------

@__all__.register
class StaticFileInterface(Interface):

    def __init__( self, root_dir = None, readonly = True, browse = True):
        self.readonly = readonly
        self.browse = browse
        self.root_dir = os.path.abspath( root_dir or '.' )
        if not os.path.isdir( self.root_dir ):
            if not self.readonly:
                raise ConfigurationError( 'StaticFileInterface: root_dir is not a directory: ' + self.root_dir )
            os.path.mkdir( self.root_dir, parents = True )

    @exposed()
    def default( self, *path, **kwargs ):
        path = path[:-1]
        path = ('/'.join(path)).replace('//', '/')
        rpath = os.path.join( self.root_dir, path )
        if os.path.isdir(rpath):
            if not self.browse:
                raise HttpNotFoundResponse( path )
            link_prefix = ( kwargs['full_url'] + '/' ).replace('//', '/')
            filelist = sorted(os.listdir(rpath))
            longest_name = reduce( lambda m, c: c if c > m else m, map(lambda n: len(n), filelist ), 0 )
            longest_name = max( longest_name, len('Parent Directory') )
            result = '<html>\n<head><title>Index of {0}</title></head>\n<body>\n<h1>Index of {0}</h1>'.format( path or '/' )
            format_row = lambda n, m, s, l = None: n + ' ' * (longest_name + 4 - (l or len(n))) + m + ' ' * (34-len(m)-len(s)) + s + '\n'
            result += '<pre>' + format_row( 'Name', 'Last Modified', 'Size' ) + '<hr>\n'
            template = '<a href="{}">{}</a>'
            format_time = lambda t: datetime.fromtimestamp(t).strftime( '%Y-%m-%d %H:%M:%S' )
            format_file = lambda l, p, n: format_row( template.format(l, n), format_time(os.path.getmtime(p)), '-' if os.path.isdir(p) else format_filesize(os.path.getsize(p)), len(n) )
            if len(path) >= 2:
                result += format_file( os.path.join(link_prefix, '../'), os.path.join(rpath,'../'), 'Parent Directory' )
            result = reduce( lambda all, name: all + format_file( os.path.join(link_prefix, name), os.path.join(rpath,name), name ), filelist, result )
            result += '</pre>\n{footer}</body>\n</html>'.format(footer=HttpResponse.footer)
            return result
        if os.path.isfile(rpath):
            return HttpFileResponse(rpath)
        raise HttpNotFoundResponse( '/' + '/'.join( kwargs['full_path'][:-1] ) )


# ---------------------------------------------------
# test
# ---------------------------------------------------

def test():
    """Start HTTP Server on port 8000, serving static files from current directory.
    """
    port = 8000
    if sys.argv[1:]:
        port = int(sys.argv[1])        
    server_address = ('', port)
    interface = StaticFileInterface(os.getcwd())
    server = Server(interface, server_address)
    server.serve_forever()

# ---------------------------------------------------
# __main__
# ---------------------------------------------------

if __name__ == '__main__':
    test()

